import re

from s_lang.data_models import Instruction, Code, Labels


def is_label_valid(label: str) -> bool:
    ch = label[0]
    idx = label[1:]
    idx = idx or "1"

    if not ch.isalpha() or not ch.isupper():
        return False

    if not idx.isnumeric():
        return False

    return 0 < int(idx)


def read_lines(code: str) -> list[str]:
    return [line.rstrip() for line in code.split("\n")]


def parse_code(code: str) -> Code:
    code = code.strip()
    code_lines: list[str] = read_lines(code)
    instructions: list[Instruction] = []
    labels: Labels = {}

    for i, line in enumerate(code_lines):
        line = line.strip()
        parts = re.split(r"^\[([A-Z]\d*)\] ", line, maxsplit=1)

        if 1 < len(parts):
            inst = parts[2]
            labels[parts[1]] = i
        else:
            inst = parts[0]

        inst_parts = re.split(r"\s+", inst)
        instructions.append(inst_parts)

    return Code(instructions, labels)



from s_lang.data_models import Labels, Instruction
from s_lang.variables_store import VariablesStore
from s_lang.parse import is_label_valid


def inc_var(variables: VariablesStore, var: str, pc: int) -> int:
    variables[var] += 1

    return pc + 1


def dec_var(variables: VariablesStore, var: str, pc: int) -> int:
    variables[var] = max(variables[var] - 1, 0)

    return pc + 1


def if_nez_goto(variables: VariablesStore, labels: Labels, var: str, label: str, pc: int) -> int:
    if not is_label_valid(label):
        raise ValueError(f"invalid label")

    val = variables[var]
    labels_target = labels.get(label, -1)

    return labels_target if 0 < val else pc + 1


def add_k(variables: VariablesStore, var: str, k: int, pc: int) -> int:
    if k < 1:
        raise ValueError("invalid const")

    variables[var] += k

    return pc + 1


def sub_k(variables: VariablesStore, var: str, k: int, pc: int) -> int:
    if k < 1:
        raise ValueError("invalid const")

    variables[var] -= k

    return pc + 1


def if_ez_goto(variables: VariablesStore, labels: Labels, var: str, label: str, pc: int) -> int:
    if not is_label_valid(label):
        raise ValueError(f"invalid label")

    val = variables[var]
    labels_target = labels.get(label, -1)

    return labels_target if 0 == val else pc + 1


def goto(labels: Labels, label: str) -> int:
    if not is_label_valid(label):
        raise ValueError(f"invalid label")

    labels_target = labels.get(label, -1)

    return labels_target


def exec_instruction(pc: int, inst: Instruction, labels: Labels, variables: VariablesStore) -> int:
    try:
        match inst:
            # basic syntax:

            case [v1, "<-", v2] if v1 == v2:
                return pc + 1

            case [v1, "<-", v2, "+", "1"] if v1 == v2:
                return inc_var(variables, v1, pc)

            case [v1, "<-", v2, "-", "1"] if v1 == v2:
                return dec_var(variables, v1, pc)

            case ["IF", v, "!=", "0", "GOTO", l]:
                return if_nez_goto(variables, labels, v, l, pc)

            # syntactic sugars:

            case [v1, "<-", v2, "+", k] if v1 == v2 and isinstance(k, str) and k.isnumeric():
                return add_k(variables, v1, int(k), pc)

            case [v1, "<-", v2, "-", k] if v1 == v2 and isinstance(k, str) and k.isnumeric():
                return sub_k(variables, v1, int(k), pc)

            case ["IF", v, "=", "0", "GOTO", l]:
                return if_ez_goto(variables, labels, v, l, pc)

            case ["GOTO", l]:
                return goto(labels, l)

            case _:
                raise ValueError("unknown command")

    except Exception as err:
        raise Exception(f"Syntax error at line {pc}: {' '.join(inst)}\n{err}")



from s_lang.data_models import Code as _Code
from s_lang.parse import parse_code as _parse_code
from s_lang.runner import Runner as _Runner

__all__ = ["run"]


def run(code: str, inputs: list[int]) -> int:
    parsed_code: _Code = _parse_code(code)
    result: int = _Runner(parsed_code, inputs).run()

    return result



from importlib import metadata

import typer

from s_lang import run as run_code


def read_file(path: str) -> str:
    with open(path, mode="r") as file:
        return file.read()


app = typer.Typer()
__version__ = metadata.version("s_lang_mta")


def version_callback(value: bool) -> None:
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def callback(
    _: bool = typer.Option(False, "--version", "-v", callback=version_callback)
) -> None:
    ...


@app.command()
def run(
    path: str = typer.Argument(help="Path to S lang code to run"),
    inputs: list[int] = typer.Argument(default=None, help="Input variables"),
) -> None:
    inputs = inputs or []
    code = read_file(path)

    result: int = run_code(code, inputs)
    typer.echo(f">>> {result}")



from collections import defaultdict


class VariablesStore:
    def __init__(self, inputs: list[int] | None = None):
        self.x: dict[int, int] = defaultdict(int)
        self.z: dict[int, int] = defaultdict(int)
        self.y: int = 0

        inputs = inputs or []

        for idx, val in enumerate(inputs, start=1):
            self.x[idx] = val

    def __getitem__(self, var: str) -> int:
        v_type, v_idx = self._parse_var(var)

        if v_type == "Y":
            return self.y

        return self._get_var_type(v_type)[v_idx]

    def __setitem__(self, var: str, value: int) -> None:
        v_type, v_idx = self._parse_var(var)

        if v_type == "Y":
            self.y = value
        else:
            self._get_var_type(v_type)[v_idx] = value

    def _get_var_type(self, v_type: str) -> dict[int, int] | None:
        if v_type == "X":
            return self.x
        elif v_type == "Z":
            return self.z

    @staticmethod
    def _parse_var(string: str) -> tuple[str, int]:
        var_char: str = string[0]
        var_idx_str: str = string[1:]

        if var_char not in ("X", "Z", "Y") or (var_char == "Y" and var_idx_str):
            raise ValueError("Invalid variable name")

        var_idx_str = var_idx_str or "1"

        if not var_idx_str.isnumeric():
            raise ValueError("Invalid variable name")

        var_idx = int(var_idx_str)

        if var_idx <= 0:
            raise ValueError("Invalid variable name")

        return var_char, var_idx



from s_lang.data_models import Code
from s_lang.variables_store import VariablesStore
from s_lang.instructions import exec_instruction


class Runner:
    def __init__(self, code: Code, inputs: list[int]):
        self.code: Code = code
        self.variables: VariablesStore = VariablesStore(inputs)
        self.pc: int = 0

    def run(self) -> int:
        prog_len = len(self.code.instructions)

        while 0 <= self.pc < prog_len:
            self.pc = exec_instruction(
                pc=self.pc,
                inst=self.code.instructions[self.pc],
                labels=self.code.labels,
                variables=self.variables
            )

        return self.variables.y
